#include <iostream>
#include <thread>
#include <atomic>
#include <vector>
using namespace std;

// Check out the following links for more information on ring buffers.
//http://www.mathcs.emory.edu/~cheung/Courses/171/Syllabus/8-List/array-queue2.html
//http://wiki.c2.com/?CircularBuffer
//https://preshing.com/20130618/atomic-vs-non-atomic-operations/

template <typename T>
class RingBuffer
{
private:
    int maxCount;
    T* buffer;
    atomic<int> readIndex;
    atomic<int> writeIndex;

public:
    RingBuffer() : maxCount(51), readIndex(0), writeIndex(0)
    {
		buffer = new T[maxCount];
        memset(buffer, 0, sizeof(buffer[0]) * maxCount);
    }

    RingBuffer(int count) : maxCount(count+1), buffer(NULL), readIndex(0), writeIndex(0)
    {
		buffer = new T[maxCount];
        memset(buffer, 0, sizeof(buffer[0]) * maxCount);
    }

    ~RingBuffer()
    {
		delete[] buffer;
		buffer = 0x0;
    }

    inline void Enqueue(T value)
    {
        // We don't want to overwrite old data if the buffer is full
        // and the writer thread is trying to add more data.  In that case,
        // block the writer thread until data has been read/removed from the ring buffer.
        while (IsFull()) { this_thread::sleep_for(500ns); }

        buffer[writeIndex] = value;
        writeIndex = (writeIndex + 1) % maxCount;
    }

    inline bool Dequeue(T* outValue)
    {
        if (IsEmpty())
            return false;

        *outValue = buffer[readIndex];
        readIndex = (readIndex + 1) % maxCount;

        return true;
    }

    inline bool IsEmpty()
    {
        return readIndex == writeIndex;
    }

    inline bool IsFull()
    {
        return readIndex == ((writeIndex + 1) % maxCount);
    }

    inline void Clear()
    {
        readIndex = writeIndex = 0;
        memset(buffer, 0, sizeof(buffer[0]) * maxCount);
    }

    inline int GetSize()
    {
        return abs(writeIndex - readIndex);
    }

    inline int GetMaxSize()
    {
        return maxCount;
    }
};

struct GfxCmd
{
public:
	virtual void Invoke() {};
};

struct GfxCmdSetRenderTarget : public GfxCmd
{
public:
	void* resourcePtr;

	GfxCmdSetRenderTarget(void* resource) : resourcePtr(resource) {}

	void Invoke()
	{
		// Invoke ID3D11DeviceContext::OMSetRenderTargets method here...
		//https://docs.microsoft.com/en-us/windows/desktop/api/d3d11/nf-d3d11-id3d11devicecontext-omsetrendertargets
		printf("%s(%p);\n", name, resourcePtr);
	}
private:
	const char* name = "GfxCmdSetRenderTarget";
};

struct GfxCmdClearRenderTargetView : public GfxCmd
{
public:
	int r, g, b;

	GfxCmdClearRenderTargetView(int _r, int _g, int _b) : r(_r), g(_g), b(_b) {}

	void Invoke()
	{
		// Invoke ID3D11DeviceContext::ClearRenderTargetView method method here...
		//https://docs.microsoft.com/en-us/windows/desktop/api/d3d11/nf-d3d11-id3d11devicecontext-clearrendertargetview
		printf("%s(%d, %d, %d);\n", name, r, g, b);

		// Pretend this command is requiring the render thread
		// to do a lot of work.
		this_thread::sleep_for(250ms);
	}
private:
	const char* name = "GfxCmdClearRenderTargetView";
};

struct GfxCmdDraw : public GfxCmd
{
public:
	int topology;
	int vertCount;

	GfxCmdDraw(int _topology, int _vertCount) : topology(_topology), vertCount(_vertCount) {}

	void Invoke()
	{
		// Invoke ID3D11DeviceContext::DrawIndexed method method here...
		//https://docs.microsoft.com/en-us/windows/desktop/api/d3d11/nf-d3d11-id3d11devicecontext-drawindexed
		printf("%s(%d, %d);\n", name, topology, vertCount);
	}
private:
	const char* name = "GfxCmdDraw";
};


void UpdateSimulationThread(RingBuffer<GfxCmd*>& gfxCmdList)
{
	// Update gameplay here.
	// Determine what to draw based on the new game state below.
	// The graphics commands will be queued up on the render thread
	// which will execute the graphics API (I.E. OpenGL/DirectX/Vulcan/etc) calls.

	gfxCmdList.Enqueue(new GfxCmdSetRenderTarget{ (void*)0x1 });
	gfxCmdList.Enqueue(new GfxCmdClearRenderTargetView{ 255, 0, 245 });
	gfxCmdList.Enqueue(new GfxCmdDraw{ 1, 10 });
}

void UpdateRenderThread(RingBuffer<GfxCmd*>& gfxCmdList)
{
	GfxCmd* gfxCmd = 0x0;
	if (gfxCmdList.Dequeue(&gfxCmd))
	{
		gfxCmd->Invoke();
		delete gfxCmd;
	}
}

void GameLoop()
{
    RingBuffer<GfxCmd*> gfxCmdList(3);
    atomic<int> counter = 0;
    atomic<bool> quit = false;

	// Run this indefinitely...
    while (1)
    {
        quit = false;
        counter = 0;
		gfxCmdList.Clear();

        thread simulationThread = thread([&gfxCmdList, &counter, &quit]
        {
			UpdateSimulationThread(gfxCmdList);
            quit = true;
        });

        thread renderThread = thread([&gfxCmdList, &quit]
        {
            // Continue to read data from the ring buffer until it is both empty
            // and the simulation thread is done submitting new items into the ring buffer.
            while (!(gfxCmdList.IsEmpty() && quit))
            {
				UpdateRenderThread(gfxCmdList);
            }
        });

        // Ensure that both the simulation and render threads have completed their work.
        simulationThread.join();
        renderThread.join();

        cout << "---\n";
    }
}

int main(int argc, char** argv[])
{
	GameLoop();

    return 0;
}